"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const exceljs_1 = require("exceljs");
const build_context_1 = require("./build-context");
const utils_1 = require("./utils");
const eval_formula_1 = require("./sandbox/eval-formula");
const debug_1 = require("./debug");
const Reference_1 = require("./coordinate/Reference");
class EmptyWorkbookError extends Error {
    constructor() {
        super(`The workbook is empty.`);
        this.name = 'EmptyWorkbookError';
    }
}
const _evalFormula = () => 0;
class Workbook extends exceljs_1.Workbook {
    normalizeCoords(label, defaultSheet) {
        return label.replace(
        // worksheet naming conventions
        // @see http://www.excelcodex.com/2012/06/worksheets-naming-conventions/
        // Additional exclusion: +-()
        /(?:'?([^\-(),:?+*'/\\]+)'?!)?\$?([a-zA-Z]+)\$?(\d+)(?::\$?([a-zA-Z]+)\$?(\d+))?/g, (label, sheet, col1, row1, col2, row2) => {
            debug_1.debug('normalizeCoords(): from %o', { label, sheet, col1, row1, col2, row2 });
            const _sheet = `${sheet ? this.getWorksheet(sheet).id : defaultSheet}!`;
            const _ref1 = `$${col1}$${row1}`;
            const _ref2 = col2 && row2 ? `:$${col2}$${row2}` : '';
            debug_1.debug('normalizeCoords():   => to "%s"', _sheet + _ref1 + _ref2);
            return _sheet + _ref1 + _ref2;
        });
    }
    async compile(spec, options = {}) {
        if (this.worksheets.length === 0)
            throw new EmptyWorkbookError();
        debug_1.debug('# Spec => %s', await debug_1.dumpToFile('spec.json', JSON.stringify(spec, (k, v) => typeof v === 'function' ? v.toString() : v, 2)));
        /**
         * MUST be serializable
         */
        const context = await build_context_1.default(Object.values(spec).map(c => Reference_1.default.from(this.normalizeCoords(c.cell, 1))), (cell, done) => {
            debug_1.debug('## Cell[%s]', cell);
            const worksheet = this.getWorksheet(cell.sheet.index.base1);
            let curCell = worksheet.getCell(cell.label);
            debug_1.debug('##   = %o', curCell.value);
            switch (curCell.type) {
                // merge cell is empty
                case 1 /* Merge */:
                    return done(null, '', '');
                // primitive values
                case 4 /* Date */:
                case 9 /* Boolean */:
                case 0 /* Null */:
                case 2 /* Number */:
                case 3 /* String */:
                    return done(null, curCell.value, curCell.value);
                // text values
                case 5 /* Hyperlink */:
                    return done(null, curCell.value.text, curCell.value.text);
                case 8 /* RichText */:
                    const richTexts = curCell.value.richText
                        .map(t => t.text)
                        .join('');
                    return done(null, richTexts, richTexts);
                // @TODO
                // SharedString is not documented (https://github.com/guyonroche/exceljs#value-types)
                // Leave it unimplemented until any reproducible scenarios
                // case ValueType.SharedString:
                case 10 /* Error */:
                    return done(null, curCell.value.error, curCell.value.error);
                case 6 /* Formula */:
                    return done(null, '=' + this.normalizeCoords(curCell.formula, cell.sheet.index.base1), curCell.result);
            }
        });
        debug_1.debug('# Context => %s', await debug_1.dumpToFile('context.json', JSON.stringify(context, null, 2)));
        const exports = {};
        for (let key of Object.keys(spec)) {
            debug_1.debug('## Exporting member[%s]', key);
            let cellType = spec[key].type ? spec[key].type
                : spec[key].args ? Function
                    : String;
            let cell = Reference_1.default.from(this.normalizeCoords(spec[key].cell, 1));
            debug_1.debug('### Exported as a %s', cellType.name);
            debug_1.debug('### Cell[%s]', cell);
            const cellValue = context[cell.toString()];
            debug_1.debug('###   = %o', cellValue);
            if (cellType === Function) {
                exports[key] = typeof cellValue === 'string' && cellValue.startsWith('=')
                    ? utils_1.wrapFunc(function (entry, args, ...evalArgs) {
                        return _evalFormula(entry, args, ...evalArgs);
                    }, {
                        entry: cell.toString(),
                        args: !spec[key].args
                            ? undefined
                            // spec[key].args! to suppress TS2532
                            // (why "!spec[key].args" does not persuade the typeguard?)
                            : spec[key].args.map(argCell => Reference_1.default.from(this.normalizeCoords(argCell, 1)).toString())
                    })
                    : utils_1.wrapFunc(function (value) {
                        return context[value];
                    }, {
                        value: cellValue
                    });
                if (options.exposeCells) {
                    exports[cell.toString()] = exports[key];
                }
            }
            else {
                exports[key] = utils_1.formatValue(cellValue, cellType);
                if (options.exposeCells) {
                    exports[cell.toString()] = exports[key];
                }
            }
        }
        const exportsFactory = utils_1.wrapFunc(function (context, exports) {
            return exports;
        }, {
            // global vars
            context,
            exports,
            _evalFormula: eval_formula_1.evalFormula
        });
        debug_1.debug('# Exports => %s', await debug_1.dumpToFile(`exports.js`, exportsFactory.toString()));
        return exportsFactory;
    }
}
exports.default = Workbook;
